# This is the template file for PSet #5, Python Task #1
import numpy,sys
import PS5_tests

class ViterbiDecoder:
    # Given the constraint length and a list of parity generator
    # functions, do the initial set up for the decoder.  The
    # following useful instance variables are created:
    #   self.K
    #   self.nstates
    #   self.r
    #   self.predecessor_states
    #   self.expected_parity
    def __init__(self, K, glist):
        self.K = K             # constraint length
        self.nstates = 2**(K-1) # number of states in state machine

        # number of parity bits transmitted for each message bit
        self.r = len(glist)

        # States are named using (K-1)-bit integers in the range 0 to
        # nstates-1. The bit representation of the integer corresponds
        # to state label in the transition diagram.  So state 10 is
        # named with the integer 2, state 00 is named with the
        # integer 0.

        # for each state s, figure out the two states in the diagram
        # that have transitions ending at state s.  Record these two
        # states as a two-element tuple.
        self.predecessor_states = \
          [((2*s+0) % self.nstates,(2*s+1) % self.nstates)
           for s in xrange(self.nstates)]

        

        # this is a 2D table implemented as a list of lists.
        # self.expected_parity[s1][s2] returns the r-bit sequence
        # of parity bits the encoder transmitted when make the
        # state transition from s1 to s2.
        self.expected_parity = \
          [[PS5_tests.expected_parity(s1, s2, K, glist) \
            if s1 in self.predecessor_states[s2] else None
            for s2 in xrange(self.nstates)]
           for s1 in xrange(self.nstates)]
        
        self.message = []

    # expected is an r-element list of the expected parity bits.
    # received is an r-element list of actual sampled voltages for the
    # incoming parity bits.  This is a hard-decision branch metric,
    # so, as described in lab write up, digitize the received voltages
    # to get bits and then compute the Hamming distance between the
    # expected sequence and the received sequence, return that as the
    # branch metric.  Consider using PS5_tests.hamming(seq1,seq2) which
    # computes the Hamming distance between two binary sequences.
    def branch_metric(self,expected,received):
        #print expected
        #print received
        #for i in xrange(recieved):
        #    PS5_tests.hamming(expected[i],received[i])
        return PS5_tests.hamming(expected,received)
        

    # Compute self.PM[...,n] from the batch of r parity bits and the
    # path metrics for self.PM[...,n-1] computed on the previous
    # iteration.  Consult the method described in the lab write up.
    # In addition to making an entry for self.PM[n,s] for each state
    # s, keep track of the most-likely predecessor for each state in
    # the self.Predecessor array (a two-dimensional array indexed by s
    # and n).  You'll probably find the following instance variables
    # and methods useful: self.predecessor_states, self.expected_parity,
    # and self.branch_metric().  To see what these mean, scan through the
    # code above.
    def viterbi_step(self,n,received_voltages):
        #is called ever time step
        #print '------------------'
        #print 'time = '+str(n)
        print received_voltages
        
        #need to convert recieved bits to ints so PS5_tests.hamming will work
        received_voltages_int = []
        for i in xrange(len(received_voltages)):
            received_voltages_int.append(int(received_voltages[i]))
        #print received_voltages_int
        
        #print self.PM

        #foreach s
        for s in xrange(self.nstates):
            #print '-----'
            #print 's: '+str(s)
            #foreach of the branchs in
            pred_length = [] #index will match with self.predecessor_states[s]
            for q in self.predecessor_states[s]:
                prev_box = self.PM[q, n-1]
                
                #print 'q: '+str(q)
                #print self.expected_parity[q][s]
                new_arrow = self.branch_metric(self.expected_parity[q][s],received_voltages_int)
                #print new_arrow
                #print prev_box
                #print prev_box+new_arrow
                pred_length.append(prev_box+new_arrow)
            #print '--'
            #check for min
            index = pred_length.index(min(pred_length))
            #print min(pred_length)
            #print self.predecessor_states[s][index]
            self.Predecessor[s,n] = self.predecessor_states[s][index]
            self.PM[s,n] = pred_length[index]
        
        #print self.Predecessor

          

    # Identify the most-likely ending state of the encoder by
    # finding the state s that has the minimum value of PM[s,n]
    # where n points to the last column of the trellis.  If there
    # are several states with the same minimum value, choose one
    # arbitrarily.  Return the tuple (s,n).
    def most_likely_state(self,n):
        #manually find min because going down a column
        #prob a better way to slice off this column
        min_state = None
        min_value = 1000000
        for s in xrange(self.nstates):
            #print s
            #print str(self.PM[s,n]) + ' vs '+ str(min_value)
            if self.PM[s,n] < min_value:
                #print 'min'
                min_state = s
                min_value = self.PM[s,n]
        return (min_state,n)


    # Starting at state s at time n, use the Predecessor
    # array to find all the states on the most-likely
    # path (in reverse order since we're tracing back through
    # the trellis).  Each state contributes a message bit.
    # Return the decoded message as a sequence of 0's and 1's.
    def traceback(self,s,n):
        #print 'traceback'
        #print 'state '+str(s)
        #print bin(s)
        #print 'which is msg: '+str(self.bit(s))
        self.message.insert(0,self.bit(s))
        #print self.message
        #print 'n is ' +str(n)

        #a little recursive action
        if (n-1) >= 1: #don't use first
            self.traceback(self.Predecessor[s,n],n-1)
        return self.message

    #oh only use first order
    #def bit(self,s):
    #    s = bin(s)       # binary representation:  bin(-37) --> '-0b100101'
    #    s = s.lstrip('-0b') # remove leading zeros and minus sign
    #    #print s
    #    #add leading 0
    #    if s == '': #empty some reason
    #        return '00'
    #    if s == '1':
    #        return '01'
    #    return s

    #return only first bit!
    def bit(self,s):
        #print bin(s)
        s = bin(s)       # binary representation:  bin(-37) --> '-0b100101'
        s = s.lstrip('-0b') # remove leading zeros and minus sign
        #print s
        #deal with leading 0
        if s == '': #empty some reason
            return 0
        if len(s) != self.K-1:
            return 0            
        return int(s[0])

    #def bit(self,s):
    #    #print self.nstates
    #    #print self.nstates/2
    #    if s <= self.nstates/2:
    #        return 0
    #    else:
    #        return 1


    # Figure out what the transmitter sent from info in the
    # received voltages.
    def decode(self,received_voltages,debug=False):
        # figure out how many columns are in the trellis
        nreceived = len(received_voltages)
        max_n = (nreceived/2) + 1

        # this is the path metric trellis itself, organized as a
        # 2D array: rows are the states, columns are the time points.
        # PM[s,n] is the metric for the most-likely path through the
        # trellis arriving at state s at time n.
        self.PM = numpy.zeros((self.nstates,max_n),dtype=numpy.float)

        # at time 0, the starting state is the most likely, the other
        # states are "infinitely" worse.
        self.PM[1:self.nstates,0] = 1000000
        #print self.PM

        # a 2D array: rows are the states, columns are the time
        # points, contents indicate the predecessor state for each
        # current state.
        self.Predecessor = numpy.zeros((self.nstates,max_n),
                                       dtype=numpy.int)

        # use the Viterbi algorithm to compute PM
        # incrementally from the received parity bits.
        n = 0
        for i in xrange(0,nreceived,self.r):
            n += 1

            # Fill in the next columns of PM, Predecessor based
            # on info in the next r incoming parity bits
            #print self.PM
            self.viterbi_step(n,received_voltages[i:i+self.r])

            # print out what was just added to the trellis state
            if debug:
                print self.PM[:,n],self.Predecessor[:,n]

        # find the most-likely ending state from the last row
        # of the trellis
        s,n = self.most_likely_state(n)
        #print 'most likely state '+str(s)+' at time '+str(n)
        #print self.PM

        # reconstruct message by tracing the most likely path
        # back through the matrix using self.Predecessor.
        return self.traceback(s,n)

    # print out final path metrics
    def dump_state(self):
        print self.PM[:,-1]

def test():
   for i in xrange(8): #from 0 to 7
        print d.bit(i)
   #sys.exit(1)

def test_hard_metrics(decoder,debug=False):
    k = 3
    glist = (7,6)
    d = decoder(k,glist)
    
    # message with no errors
    print 'message with no errors'
    message = numpy.random.random_integers(0,1,10)
    sent = PS5_tests.convolutional_encoder(message,k,glist)
    received_message = d.decode(sent,debug=debug)
    if numpy.any(message != received_message):
        print "Decoder testing failed (no transmission errors)..."
        print "Received parity bits:",sent
        print "Decoded message returned by your decoder: ",received_message
        print "Expected:",message
        return

    
    d = decoder(k,glist)
    # message with two transmission errors
    print 'message with two errors'
    sent[3] = 1 - sent[3]
    sent[12] = 1 - sent[12]
    received_message = d.decode(sent,debug=debug)
    if numpy.any(message != received_message):
        print "Decoder testing failed (2 transmission errors)..."
        print "Received parity bits:",sent
        print "Decoded message returned by your decoder: ",received_message
        print "Expected:",message
        return

    print "Hard-metric Viterbi decoder tests complete!"

if __name__=='__main__':
    K = 3; glist = (7,6)
    d = ViterbiDecoder(K, glist)

    # first test case: example from lecture
    #that is not the message from the lecture!
    message = [1,0,1,0,1,1,0,0] #p-set native
    #message = [1,0,1,1,0,0] #and then glist also does not work
    
    received = PS5_tests.convolutional_encoder(message, K, glist)

    i = 0
    print '==============================================='
    print 'TEST', i
    print message
    print received
    decoded = numpy.array(d.decode(received, debug=True))
    print 'Testing without adding noise...'
    if (message == decoded).all() == True: 
        print 'Successfully decoded no-noise Test 0: congratulations!'
        print
    else:
        print 'Oops... error in decoding no-noise Test', i
        print 'Decoded as ', decoded
        print 'Correct is', message
        #sys.exit(1)

    # second batch of test cases: different constraint lengths, generators
    nbits = 29
    message = numpy.random.random_integers(0,1,nbits)
    for (K, glist) in ((3, (7,6)), (4, (0xD,0xE))):
        i = i + 1
        print '==============================================='
        print 'TEST', i
        print message
        d = ViterbiDecoder(K, glist)
        received = PS5_tests.convolutional_encoder(message, K, glist)
        print received
        decoded = numpy.array(d.decode(received, debug=True))
        if (message == decoded).all() == True: 
            print 'Successfully decoded no-noise Test', i, ': congratulations!'
            print
        else:
            print 'Oops... error in decoding no-noise Test', i
            print 'Decoded as', decoded
            print 'Correct is', message
            #sys.exit(1)


    # now try some tests with noise
    #NOTE: I modified the test to reinitalize the decoder state upon reuse
    #I just added: d = decoder(k,glist)
    test_hard_metrics(ViterbiDecoder, True)
    
