import math,cmath,numpy,os,wave
import matplotlib.pyplot as p
import matplotlib.ticker as ticker

# read in .wav file data, convert to sequence of floats
# scaled between -1 and 1
def read_wav_file(filename):
    assert os.path.exists(filename),\
           "file %s doesn't exist" % filename
    wav = wave.open(filename,'rb')
    total_frames = wav.getnframes()
    nchan = wav.getnchannels()
    assert total_frames > 0,\
           "%s doesn't have any audio data!" % self.filename
    sample_rate = wav.getframerate()
    sample_width = wav.getsampwidth()

    # see http://ccrma.stanford.edu/courses/422/projects/WaveFormat/
    if sample_width == 1:
        # data is unsigned bytes, 0 to 255
        dtype = numpy.uint8
        scale = 1/127.0
        offset = -1.0
    elif sample_width == 2:
        # data is signed 2's complement 16-bit samples (little-endian byte order)
        dtype = numpy.int16
        scale = 1/ 32767.0
        offset = 0.0
    elif sample_width == 4:
        # data is signed 2's complement 16-bit samples (little-endian byte order)
        dtype = numpy.int32
        scale = 1/2147483647.0
        offset = 0.0
    else:
        assert False,"%s: unrecognized sample width %d" % (filename,sample_width)

    # read in data
    frames = wav.readframes(total_frames)
    wav.close()

    # only use 1st channel
    signal = numpy.frombuffer(frames,dtype=dtype)
    signal = signal[::nchan]

    # scale data appropriately
    signal = signal*scale + offset

    return list(signal)


# pretty plot of a sequence
def plot(subplot,x,y,title):
    if isinstance(subplot,list):
        p.subplot(subplot[0],subplot[1],subplot[2])
    else:
        p.subplot(subplot)
    p.subplots_adjust(top=0.9,hspace=0.8)
    p.stem(x,y,linefmt='b-',markerfmt='bo',basefmt='k-')
    p.grid(True)
    smin = min(y)
    smax = max(y)
    ds = smax - smin
    if ds < 1e-10: ds = .001
    p.axis([x[0] - .5,x[0] + len(x) - .5,smin - .1*ds,smax + .1*ds])
    p.title(title,fontsize=14,va='bottom')

# plot the real and imaginary part of spectral coefficients
def plot_spectrum(a):
    p.figure()
    N = len(a)
    halfN = N/2
    x = [i - halfN for i in xrange(N)]
    plot(211,x,[y.real for y in a],r'$\Re(a_k)$')
    plot(212,x,[y.imag for y in a],r'$\Im(a_k)$')

# radians formatting
def rad_fn(x,pos=None):
    y = 0 if x == 0 else int(4*x/math.pi)
    if y < 0:
        sign = '-'
        y = -y
    else:
        sign = ''
    if y == 0:
        return '0'
    elif y == 4:
        return sign+'$\pi$'
    elif y == 2:
        return sign+'$\pi/2$'
    elif y == 1:
        return sign+'$\pi/4$'
    else:
        return sign+'$%d\pi/4$' % y

# plot the frequency response of an LTI system given its
# unit sample response
def plot_freq_response(h):
    N = 200
    omegas = numpy.linspace(-math.pi,math.pi,N)
    H = numpy.zeros(N,dtype=complex)
    for i,omega in enumerate(omegas):
        for m in xrange(len(h)):
            H[i] += h[m]*cmath.exp(complex(0,-omega*m))
    p.figure()
    ax = p.subplot(111)
    absH = abs(H)
    p.plot(omegas,absH)
    p.xlim(-3.2,3.2)
    p.ylim(-.1*max(absH),1.1*max(absH))
    ax.xaxis.set_major_locator(ticker.MultipleLocator(math.pi/4))
    ax.xaxis.set_major_formatter(ticker.FuncFormatter(rad_fn))
    p.grid(True)
    p.title(r'$|H(e^{j\Omega})|$')
        

